<div>
  <sly-apexchart
    style="margin-bottom: 15px"
    :options="data.{{{ widget.widget_id }}}.options"
    :series="data.{{{ widget.widget_id }}}.series"
    :type="data.{{{ widget.widget_id }}}.type"
    :height="data.{{{ widget.widget_id }}}.height"
    :sly-options="data.{{{ widget.widget_id }}}.sly_options"
    {%
    if
    widget._click_handled
    %}
    @click="state.{{{ widget.widget_id }}}.clicked_value = $event;
            post('/{{{ widget.widget_id }}}/chart_clicked_cb')"
    {%
    endif
    %}
  />
</div>
